/* -*- Mode: C -*- */

//   $Id$

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <assert.h>
#include <avr/io.h>

#include "gaul.h"

  // --------------------------------------------------------------------------
  // DEFINITIONS 
  // --------------------------------------------------------------------------

#define PSU_STATUS_ID_PSU_OFF             (1<<0)
#define PSU_STATUS_TEXT_PSU_OFF           "OFF"

#define PSU_STATUS_ID_PSU_BOOTING         (1<<1)
#define PSU_STATUS_TEXT_PSU_BOOTING       "BOOT"

#define PSU_BUFFER_LENGTH                    128

#define PSU_IO_UPDATE_INTERVAL              0.01  // [s]
#define PSU_IO_TIMER_PRESCALER_VALUE           1

#define HEART_BEAT_TICKS                    5000

#define DARC_CMD_1                         "SAY-HELLO"
#define DARC_CMD_2                         "SAY-GOODBYE"

  // --------------------------------------------------------------------------
  // DATA TYPES 
  // --------------------------------------------------------------------------

typedef uint16_t tStatusID;

  // --------------------------------------------------------------------------
  // GLOBAL VARS 
  // --------------------------------------------------------------------------

volatile uint16_t  nTimerLoadValue;
char acBuffer[ PSU_BUFFER_LENGTH ];

  // --------------------------------------------------------------------------
  // FUNCTIONS
  // --------------------------------------------------------------------------

void init_port_b ( void )
{
  uint8_t nTmpSREG = SREG; // Save status 

  cli();                   // Ensure atomic execution

  DDRB |= (1<<PB0);        // Set line 0 of port B as output
  DDRB |= (1<<PB1);        // Set line 1 of port B as output
  DDRB |= (1<<PB2);        // Set line 2 of port B as output
  DDRB |= (1<<PB3);        // Set line 3 of port B as output
  DDRB |= (1<<PB4);        // Set line 4 of port B as output

  // Heartbeat LED: Line 0 -> Switch off initially

  PORTB &= ~( 1<<PB4 );

  asm volatile ( "nop" );  // Give the AVR time to settle

  SREG = nTmpSREG;         // We're done so interrupts ok now 

  return;
}

void get_status_text( tStatusID nStatus, char **ppcBuffer )
{
  if( *ppcBuffer == NULL )
    return;

  switch( nStatus )
  {
    case PSU_STATUS_ID_PSU_OFF:
      *ppcBuffer = strcpy_P( (char *) &acBuffer[0], PSTR( PSU_STATUS_TEXT_PSU_OFF ));
      break;

    case PSU_STATUS_ID_PSU_BOOTING:
      *ppcBuffer = strcpy_P( (char *) &acBuffer[0], PSTR( PSU_STATUS_TEXT_PSU_BOOTING ));
      break;

    default:
      *ppcBuffer = NULL;
      break;
  }
}

void init_io_timer ( void )
{
  double   nTickInterval;
  double   nDesiredTimerInterval;
  uint16_t nPrescalerValue;

  uint8_t  nTmpSREG = SREG;      // Save status 
 
  // We use timer 1 (16 bit Timer) in CTC mode. This timer shall 
  // generate an interrupt every PSU_IO_UPDATE_INTERVAL seconds.
  // Prescaler has to be set to 1 !

  cli();                         // Ensure atomic execution

  nTickInterval         = 0.0;   // [s]
  nDesiredTimerInterval = PSU_IO_UPDATE_INTERVAL;
  nPrescalerValue       = PSU_IO_TIMER_PRESCALER_VALUE;

  nTickInterval   = 1 / F_CPU * nPrescalerValue;
  nTimerLoadValue = nDesiredTimerInterval / nTickInterval - 1; 
    
  OCR1A  = nTimerLoadValue;      // Load the timer compare value
  
  TCNT1  = 0;                    // Reset counter value
  TCCR1A = 0;                    // No PWM
  TIMSK  |= _BV( OCIE1A );       // Output Compare A Interrupt: enable it
  TCCR1B |= _BV( WGM12 );        // Set CTC mode
  
  TCCR1B &= ~( _BV( CS12 ));     // Set prescaler to 1
  TCCR1B &= ~( _BV( CS11 ));     // and start timer.
  TCCR1B |= _BV( CS10 );         

  SREG = nTmpSREG;               // Done - so interrupts ok now
}

void heart_beat( void )
{
  static uint32_t nTicks = 0;
    
  nTicks++;

  if( nTicks > HEART_BEAT_TICKS )
  {
    PORTB ^= ( 1 << PB4 );
    nTicks = 0;
  }
}

ISR( TIMER1_COMPA_vect  )    // ISR triggered by Timer 1 to
{                            // interact with the user via DARC.

  heart_beat();              // Blink heartbeat LED.

  darc_get_command();        // Receive cmd via DARC channel. Will set
                             // a flag if complete cmd received.
}

// --------------------------------------------------------------------------
// M A I N
// --------------------------------------------------------------------------

int main ( void )
{
  tsDARCEvent sEvent;
  
  char acBuf1[ 21 ];
  char acBuf2[ 21 ];

  bool_t bCmdOk;
  char *pcCmd = NULL;

  lcd_init( LCD_DISP_ON, 0xc6 ); // Init LCD at I2C address 0xC6

  // Port B: Heartbeat LED

  init_port_b();

  // Enable error logging in EEPROM

  init_error_logging();
  show_last_logged_error( 1000 );

  // DARC = Digital Appliance Remote Control (UART-based comm interface)
  // UART comm is done using Peter Fleury's UART Library.
  // (see http://jump.to/fleury)

  enable_darc();
  darc_init( 38400 ); // Comm params: 38400 Baud 8 Bits No Parity 1 Stop Bit

  // Globally enable interrupts

  sei();

  // Setup periodic I/O update timer

  init_io_timer(); // DARC send works if Timer 1 is disabled by
                   // commenting out this line...

  // Prepare LCD
  
  clear_display();

  // Announce device startup on DARC channel

  sEvent.nType = DARC_EVENT_TYPE_BOOT;
  sEvent.pcInfoText1 = strcpy_P( (char *) &acBuf1[0], PSTR( "P1025A" ));
  sEvent.pcInfoText2 = strcpy_P( (char *) &acBuf2[0], PSTR( "801" ));
  darc_send_event( &sEvent );

  sEvent.nType = DARC_EVENT_TYPE_STATUS_CHANGE;
  sEvent.pcStatusText = &acBuffer[ 0 ];
  get_status_text( PSU_STATUS_ID_PSU_BOOTING, &( sEvent.pcStatusText ));
  darc_send_event( &sEvent );

  // Check for command received via DARC channel.
  // Receiving is done in ISR for I/O timer via function
  // darc_get_command(). This function sets a flag indicating a 
  // full DARC command has been received. If command is not received
  // completely then function exits as soon as no data is available
  // in UART receive ring buffer (= "standard" behaviour of Peter
  // Fleury's UART Library).
  
  // Debugging loop ...
  for(;;)
  {  
    // darc_get_command(); // Should really be done in ISR of Timer 1 !!!

    // DARC command received ?
    if( is_darc_command_available() )
    {
      // Yes! So, for debugging purposes, we show this on the LCD.
      lcd_gotoxy( 4, 1 );
      lcd_puts( "Cmd: " );
      pcCmd = strupr( darc_receive_buffer() );
      lcd_puts( pcCmd );

      // Simple debugging ...
 
      bCmdOk = FALSE;
 
      if( ! strncmp( pcCmd, DARC_CMD_1, strlen( DARC_CMD_1 )))
      {
        bCmdOk = TRUE;
        darc_send_result( RC_OK );
        darc_format( "HELLO, DEAR !\r\n" );
      }

      if(( ! bCmdOk ) && (! strncmp( pcCmd, DARC_CMD_2, strlen( DARC_CMD_2 ))))
      {
        bCmdOk = TRUE;
        darc_send_result( RC_OK );
        darc_format( "GOODBYE ...\r\n" );
      }
   
      if( bCmdOk == FALSE )
        darc_send_result( RC_ERROR );
          
      // Reset DARC command receiving thereby enabling the reception 
      // of a new command.
      darc_clear_command_receiving();

      // Wait a second ...
      _delay_ms( 1000 );
   
      // Cleanup LCD
      lcd_clrscr();
    }
    else
    {
      lcd_home();
      lcd_puts("> ");
      lcd_puts( darc_receive_buffer() );
    }
  }

  return 0;
}
