/* -*- Mode: C -*- */

// ===========================================================================
//
//   --->>>   G A U L   <<<---
//
//   * G oenninger
//   * A VR
//   * U tility
//   * L ibrary
//
//   This file:   D A R C . C
//                Digital Appliance Remote Control
//
//   $Id: darc.c,v 1.1 2009/02/01 17:07:22 frgo Exp frgo $
//
// ===========================================================================

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>

#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "gaul.h"

// --------------------------------------------------------------------------
// GLOBAL DEFINITIONS
// --------------------------------------------------------------------------

// DEBUG "SWITCHES"

// --------------------------------------------------------------------------
// GLOBAL VARS
// --------------------------------------------------------------------------

static FILE darc_stdio = 
  FDEV_SETUP_STREAM( uart_putc, NULL, _FDEV_SETUP_WRITE );

static volatile FILE *_darc_stream;

static volatile bool_t bUseDARC = FALSE;
static volatile bool_t bDARCIsInitialized;

static char acDARCReceiveBuffer[ DARC_RECEIVE_BUFFER_LENGTH + 1 ];
static volatile uint32_t nDARCReceiveBufferIndex;

static char acDARCSendBuffer[ DARC_SEND_BUFFER_LENGTH + 1 ];
static volatile uint32_t nDARCSendBufferIndex;

static volatile bool_t bDARCCommandAvailable;

// --------------------------------------------------------------------------
// DARC I/O
// --------------------------------------------------------------------------

void enable_darc( void )
{
  bUseDARC = TRUE;
}

void disable_darc( void )
{
  bUseDARC = FALSE;
}

bool_t use_darc( void )
{
  return bUseDARC;
}

bool_t is_darc_command_available( void )
{
  return bDARCCommandAvailable;
}

void clear_darc_command_is_available( void )
{
  bDARCCommandAvailable = FALSE;
}

FILE *darc_stream( void )
{
  return _darc_stream;
}

void darc_clear_receive_buffer( void )
{
  acDARCReceiveBuffer[ 0 ] = '\0';
  nDARCReceiveBufferIndex = 0;
}

char *darc_receive_buffer( void )
{
  return acDARCReceiveBuffer;
}

void darc_clear_send_buffer( void )
{
  acDARCSendBuffer[ 0 ] = '\0';
  nDARCSendBufferIndex = 0;
}

rc_t ensure_darc_init( void ) 
{
  if( bDARCIsInitialized == FALSE )
    return darc_init( DARC_BAUDRATE );
  else
    return RC_OK;
}

rc_t darc_init( unsigned long nBaudrateSelector )
{
  rc_t nRC = RC_BUSY;

  if( bUseDARC )
  {
    // Initialize UART of the controller board (uses Peter Fleury's
    // UART Library).

    uart_init( UART_BAUD_SELECT( 38400, F_CPU ));

    // Set streams

    stdout = (FILE *) &darc_stdio;
    // stdin  = (FILE *) &darc_stdio;

    _darc_stream = (FILE *) &darc_stdio;
   
    // Init buffers
 
    darc_clear_receive_buffer();
    darc_clear_send_buffer();
    
    // Save init status
 
    bDARCIsInitialized    = TRUE;
    bDARCCommandAvailable = FALSE;

    // Send string of spaces for syncing ...

    darc_format( "   \r\n" );

    // Done.

    nRC = RC_OK;
  }

  return nRC;
}

int darc_putc( unsigned char c, FILE *stream )
{
  uart_putc( c );
   
  return 0;
}

int darc_puts( const char * pcString )
{
  uart_puts( pcString );

  return 0;
}

unsigned int darc_getc( FILE *stream )
{
  return uart_getc();
}

rc_t darc_format( const char *pcFormat, ... )
{
  rc_t nRC = RC_OK;
  
  if( bUseDARC && ( ensure_darc_init() == RC_OK ))
  {   
    va_list argp;
    
    va_start( argp, pcFormat );

    vsnprintf( acDARCSendBuffer, 
               DARC_SEND_BUFFER_LENGTH + 1,
               pcFormat,
               argp );

    /*** Doesn't work:
    vfprintf( _darc_stream, 
              pcFormat, 
              argp );
    ***/

    va_end( argp ); 

    darc_puts( acDARCSendBuffer );
    darc_clear_send_buffer();
  }

  return nRC;
}

rc_t darc_receive( void ) // Can be called repeatedly until a full command 
{                         // has been received !
  rc_t nRC = RC_BUSY;

  unsigned int   cDARCChar  = '\0';
  static bool_t  bDone      = TRUE;

  while( 1 )
  {
    cDARCChar = uart_getc();

    // When no data received or frame error then command is not fully 
    // received yet => return from darc_receive with Done flag set to false.

    if(( cDARCChar & UART_NO_DATA ) ||
       ( cDARCChar & UART_FRAME_ERROR ))
    { 
      bDone = FALSE;
      break;
    }    
    
    // Save received char in receive buffer

    acDARCReceiveBuffer[ nDARCReceiveBufferIndex++ ] = 
      (unsigned char) cDARCChar;
    acDARCReceiveBuffer[ nDARCReceiveBufferIndex ] = '\0';
 
    // Check if received char marks command or request end

    if(( cDARCChar == DARC_CAT_CMD_END_CHAR ) || 
       ( cDARCChar == DARC_CAT_CMD_REQUEST_END_CHAR )) 
    {
      bDone = TRUE;
      break;
    }

    // At end of space in receive buffer ? If so, exit

    if( nDARCReceiveBufferIndex >= DARC_CAT_CMD_MAX_LENGTH )
    {
      bDone = TRUE;
      break;
    }
  }

  // Commnd fully received ? If so then set flag.
  // If not indicate this via RC_BUSY.

  if( bDone == TRUE )
  {
    bDARCCommandAvailable = TRUE;
    nRC = RC_OK;
  }
  else
    nRC = RC_BUSY;

  return nRC;
}

rc_t darc_send_result( rc_t nResultRC )
{
  rc_t nRC = RC_BUSY;

  if( bUseDARC && ( ensure_darc_init() == RC_OK ))
  {   
    char *pcResultString = NULL;

    switch( nResultRC )
    {
      case RC_OK:

        pcResultString = DARC_RS_OK;
        break;

      default:

        pcResultString = DARC_RS_FAIL;
        break;
    }
  
    if( pcResultString != NULL )
    {
      nRC = darc_format( pcResultString );    
    }
  }

  return nRC;  
}

rc_t darc_send_event( tpsDARCEventPtr psEvent )
{
  rc_t nRC = RC_OK;

  if( bUseDARC && ( ensure_darc_init() == RC_OK ))
  {   
    if( psEvent != NULL )
    {
      switch( psEvent->nType )
      {
        case DARC_EVENT_TYPE_FAILURE_MESSAGE:
          nRC = darc_format( "EVENT FAILURE REASON=%d.\r\n",
                             psEvent->nFailureReasonCode );
          break;

        case DARC_EVENT_TYPE_STATUS_CHANGE:
          nRC = darc_format( "EVENT STATUS-CHANGE STATUS=%s.\r\n",
                             psEvent->pcStatusText );
          break;

        case DARC_EVENT_TYPE_BOOT:
          nRC = darc_format( "EVENT BOOT DEVICE=%s ID=%s.\r\n",
                             psEvent->pcInfoText1,
                             psEvent->pcInfoText2 );
          break;

        default:
          break;
      }
    }
  }

  return nRC;
}

void darc_clear_command_receiving( void )
{
  // Reset DARC thereby enabling the reception of a new command.
  darc_clear_receive_buffer();
  clear_darc_command_is_available();
}

void darc_get_command( void )
{
  if( ! bUseDARC )
    return;
  
  // Already a command received ? If flag is set then DARC still needs
  // to process the previously received command ...

  if( bDARCCommandAvailable == FALSE ) 
  {
    darc_receive();
  } 
}
