/* -*- Mode: C -*- */

#include "lcd.h"
#include "lcd_tools.h"
#include "display.h"
#include "error.h"
#include "eeprom.h"

// --------------------------------------------------------------------------
// LOCAL DEEINES                                                              
// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
// Special VARS                                                              
// --------------------------------------------------------------------------

static volatile uint16_t nErrorNr        = 0;
static volatile uint16_t nErrorLine      = 0;

static volatile uint16_t nLastErrorNr    = 0;
static volatile uint16_t nLastErrorLine  = 0;

static volatile uint16_t nNrLoggedErrors = 0;

// --------------------------------------------------------------------------
// INIT_ERROR_LOGGING                                                   
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   None
//
// --------------------------------------------------------------------------

void init_error_logging ( void )
{
  get_last_logged_error();
  
  nErrorNr = 0;
  nErrorLine = 0;
  
  nLastErrorNr = 0;
  nLastErrorLine = 0;
  
  return;
}

// --------------------------------------------------------------------------
// RESET_EEPROM_ERROR_LOG                                                   
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   None
//
// --------------------------------------------------------------------------

void reset_eeprom_error_log ( void )
{
   
}

// --------------------------------------------------------------------------
// GET_NR_LOGGED_ERRORS                                                   
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   None
//
// --------------------------------------------------------------------------

uint16_t get_nr_logged_errors ( void )
{
  return nNrLoggedErrors;
}

// --------------------------------------------------------------------------
// LOG_ERROR                                                    
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   nErrorNr
//   nErrorLine
//
// --------------------------------------------------------------------------

void log_error ( uint16_t _nErrorNr, 
								 uint16_t _nErrorLine )
{
  uint16_t   nTmpSREG  = 0;
  uint8_t    nTries    = 0;
  uint16_t  *pvAddress = (uint16_t *) NULL;

  nErrorNr   = _nErrorNr;
  nErrorLine = _nErrorLine;

  if( nNrLoggedErrors > ( EEPROM_MAX_LOGGED_ERRORS - 1 ))
	{
    reset_eeprom_error_log();
	}

  incf_nr_logged_errors();

  nTmpSREG = SREG;

  cli(); // No interrupts please ...

  // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

  while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		     ( ! eeprom_is_ready() ))
	{
    nTries++;
    _delay_ms( EEPROM_WAIT_TIME );
  }

  // Calc address and write error and then error line

  pvAddress = 
		pvEEMEM + sizeof( uint16_t ) + 
		( nNrLoggedErrors - 1)  *  ( 2 * sizeof( uint16_t ));

  eeprom_write_word( pvAddress, nErrorNr );

  pvAddress += sizeof( uint16_t );
  eeprom_write_word( pvAddress, nErrorLine);

  // Enable interupts 
  
  SREG = nTmpSREG;

  return;
}

// --------------------------------------------------------------------------
// READ_NR_LOGGED_ERRORS                                                    
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   None
//
// --------------------------------------------------------------------------

void read_nr_logged_errors ( void )
{
  uint16_t nTmpSREG  = SREG;
  uint8_t  nTries    = 0;
 
  cli(); // No interrupts please ...

  // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

  while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		     ( ! eeprom_is_ready() ))
	{
    nTries++;
    _delay_ms( EEPROM_WAIT_TIME );
  }
  
  // Read nr of logged errors

  nNrLoggedErrors = eeprom_read_word( pvEEMEM );

  // Enable interrupts
  
  SREG = nTmpSREG;  

  if ( nNrLoggedErrors == -1 ) // true when AVR just has been programmed
    incf_nr_logged_errors();
 
  return;    
}

// --------------------------------------------------------------------------
// INCF_NR_LOGGED_ERRORS                                                    
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   None
//
// --------------------------------------------------------------------------

void incf_nr_logged_errors ( void )
{
  uint16_t nTmpSREG  = SREG;
  uint8_t  nTries    = 0;
 
  cli(); // No interrupts please ...

  // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

  while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		     ( ! eeprom_is_ready() ))
	{
    nTries++;
    _delay_ms( EEPROM_WAIT_TIME );
  }
  
  // Increment nr logged errors

  nNrLoggedErrors++;

  // Write new nr of logged errors

  eeprom_write_word( pvEEMEM, nNrLoggedErrors );
  
  // Enable interrupts
  
  SREG = nTmpSREG;  
 
  return;    
}

// --------------------------------------------------------------------------
// GET_LAST_LOGGED_ERROR                                                    
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   & Error Info (2 uint16_t pointers)
//
// --------------------------------------------------------------------------

void get_last_logged_error ( void )
{
  uint16_t nTmpSREG   = SREG;
  uint8_t  nTries     = 0;
  uint16_t *pvAddress = (uint16_t *) NULL;

  read_nr_logged_errors();

  if( nNrLoggedErrors > 0 )
  {
    cli(); // No interrupts please ...

    // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

    while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		       ( ! eeprom_is_ready() ))
	  {
      nTries++;
      _delay_ms( EEPROM_WAIT_TIME );
    }

    // Calc address of last logged error and then read error nr info 

    pvAddress = pvEEMEM + sizeof( uint16_t ) + 
		  ( nNrLoggedErrors - 1 ) * ( 2 * sizeof( uint16_t ));
  
    nLastErrorNr = eeprom_read_word( pvAddress );

    // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

    nTries = 0;

    while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		       ( ! eeprom_is_ready() ))
	  {
      nTries++;
      _delay_ms( EEPROM_WAIT_TIME );
    }

    // Read error line info

    pvAddress += sizeof( uint16_t );  

    nLastErrorLine = eeprom_read_word( pvAddress );
  
    // Enable interrupts
  
    SREG = nTmpSREG;  

  }
  else  // no error logged so far ...
  {
    nNrLoggedErrors = 0;
		nLastErrorNr    = 0;
		nLastErrorLine  = 0;
  }

  return;     
}

// --------------------------------------------------------------------------
// SHOW_LAST_LOGGED_ERROR                                                    
// --------------------------------------------------------------------------

void show_last_logged_error ( unsigned long nDelayMilliSeconds )
{
  char acBuffer[ LCD_LINE_LENGTH + 1 ];
  
  memset( acBuffer, 0, LCD_LINE_LENGTH + 1 );
  
  clear_display();  
  get_last_logged_error();

  display_line( 1, "LAST LOGGED ERROR:" );
  if ( nNrLoggedErrors == 0 )
    display_line( 2, "***  None  ***" );
  else
  {
    snprintf( acBuffer,
              LCD_LINE_LENGTH,
              "* %03d: %04d-%04d *" , 
						 (int) nNrLoggedErrors, (int) nLastErrorNr, (int) nLastErrorLine );
    display_line( 2, acBuffer );
  }

  if( nDelayMilliSeconds > 0 )   
    _delay_ms( nDelayMilliSeconds );

  return;     
}

// --------------------------------------------------------------------------
// GET_ERROR_NR                                                    
// --------------------------------------------------------------------------

uint16_t get_error_nr ( void )
{
  return nErrorNr;
}

// --------------------------------------------------------------------------
// GET_ERROR_LINE                                                    
// --------------------------------------------------------------------------

uint16_t get_error_line ( void )
{
  return nErrorLine;
}

// --------------------------------------------------------------------------
// GET_LOGGED_ERROR                                                    
// --------------------------------------------------------------------------

rc_t get_logged_error ( uint16_t nIndex, uint16_t *pnThisNr, uint16_t *pnThisLine )
{
  uint16_t nTmpSREG   = SREG;
  uint8_t  nTries     = 0;
  uint16_t *pvAddress = (uint16_t *) NULL;

  if(( pnThisNr == NULL ) || ( pnThisLine == NULL ))
  {
    return SW_ERROR;
  }

  if( nNrLoggedErrors > 0 )
  {
    cli(); // No interrupts please ...

    // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

    while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		       ( ! eeprom_is_ready() ))
	  {
      nTries++;
      _delay_ms( EEPROM_WAIT_TIME );
    }

    // Calc address of last logged error and then read error nr info 

    pvAddress = pvEEMEM + sizeof( uint16_t ) + 
		  nIndex * ( 2 * sizeof( uint16_t ));
  
    *pnThisNr = eeprom_read_word( pvAddress );

    // Wait EEPROM_MAX_WAIT_TRIES for EEPROM to become ready

    nTries = 0;

    while (( ! ( nTries == EEPROM_MAX_WAIT_TRIES ) ) &&
		       ( ! eeprom_is_ready() ))
	  {
      nTries++;
      _delay_ms( EEPROM_WAIT_TIME );
    }

    // Read error line info

    pvAddress += sizeof( uint16_t );  

    *pnThisLine = eeprom_read_word( pvAddress );
  
    // Enable interrupts
  
    SREG = nTmpSREG;  
  }

  return RC_OK;     
}

// --------------------------------------------------------------------------
// PRINT_ERROR                                                    
// --------------------------------------------------------------------------
//                                                                           
// Parameters:
//   stream    stream on which to output the XML-encoded error info
//
// --------------------------------------------------------------------------

rc_t print_errorlog( FILE *stream )
{
  rc_t nRC = RC_OK;

  uint16_t nThisNr   = 0;
  uint16_t nThisLine = 0;

  if( stream == NULL )
  {
    log_error( SW_ERROR, __LINE__ );
    return SW_ERROR;
  }
 
  read_nr_logged_errors(); // Make sure we have an up-to-date nr of errors
 
  if( nNrLoggedErrors > 0 )
  {
    fprintf( stream, "<errorlog nr_errors=%d>", nNrLoggedErrors );
    
    for( uint16_t nI = 0; nI < nNrLoggedErrors; nI++ )
    {
      nRC = get_logged_error( nI, &nThisNr, &nThisLine );
      if( nRC == RC_OK )
      {  
        fprintf( stream, "<error index=%d>", nI );
        fprintf( stream,   "<nr>%ul</nr>", nThisNr );
        fprintf( stream,   "<line>%ul</line>", nThisLine );
        fprintf( stream, "</error>" );
      }
    }
  }
  
  return nRC;
}
